Page
Library
Module
Module type
Parameter
Class
Class type
PbSourceField types
The type of protobuf field encodings.
Varint-encoded booleans
Varint-encoded signed 32-bit integers. This is space-inefficient for negative numbers; it is recommended to use sint32 instead if negative numbers are likely to be frequent.
Zigzag-encoded signed 32-bit integer. This is more space-efficient for negative numbers than int32.
Fixed-size encoding of signed 32-bit integers.
Varint-encoded signed 64-bit integers. This is space-inefficient for negative numbers; it is recommended to use sint64 instead if negative numbers are likely to be frequent.
Zigzag-encoded signed 64-bit integer. This is more space-efficient for negative numbers than int64.
Fixed-size encoding of signed 64-bit integers.
Varint-encoded unsigned 32-bit integers.
Fixed-size encoding of unsigned 32-bit integers.
Varint-encoded unsigned 64-bit integers.
Fixed-size encoding of unsigned 64-bit integers.
Fixed-size encoding of 64-bit floating-point numbers.
Fixed-size encoding of 32-bit floating-point numbers.
Length-delimited encoding of UTF-8 strings.
Length-delimited encoding of byte strings.
A value of type msg represents a message
A value of type msgtype represents a message type
Embedded message field type
read_field t returns an Angstrom parser for a field descsribed by t.
write_field t f v serializes the bytes of the encoding of v described by t, in order, to f.
A value of type (m, t) field describes a field of type t in a message of type m.
Pretty-print a field type
Operations on values
Pretty-print a field value
Pretty-print a message
read mt returns an Angstrom parser for a message described by mt.